from enum import Enum, auto
from dataclasses import dataclass
from datetime import datetime


class EventType(Enum):
    STATUS = auto()
    DUE_DATE = auto()
    NAME = auto()
    PROJECT = auto()
    DESCRIPTION = auto()
    PRIORITY = auto()
    ASSIGNEE = auto()
    ADDED_TO_SPRINT = auto()
    REMOVED_FROM_SPRINT = auto()
    ADDED_TO_EPIC = auto()
    REMOVED_FROM_EPIC = auto()
    ADDED_TO_STORY = auto()
    REMOVED_FROM_STORY = auto()


@dataclass
class TaskEvent:
    task: int
    type: str
    old_value: str
    new_value: str
    date: datetime = datetime.now()

    def __post_init__(self) -> None:
        try:
            EventType[self.type]
        except KeyError as e:
            raise ValueError(
                f"{self.type} is invalid EventType for Task") from None
        # TODO: Select better name
        self.syncable = self.type in ("STATUS", "DUE_DATE", "NAME",
                                        "PROJECT", "DESCRIPTION", "PRIORITY")


@dataclass
class ProjectEvent:
    project: int
    event_type: str
    old_value: str
    new_value: str
    date: datetime = datetime.now()

    def __post_init__(self) -> None:
        if self.event_type.upper() not in [e.name for e in EventType]:
            raise ValueError(f"{self.event_type} is invalid EventType")



import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, clear_mappers

from terka import bootstrap
from terka.adapters.orm import metadata, start_mappers
from terka.adapters import publisher, repository
from terka.domain import events
from terka.service_layer import unit_of_work


class FakePublisher(publisher.BasePublisher):

    def __init__(self):
        self.events = []  # type: Dict[str, List[str]]

    def publish(self, topic, event):
        self.events.append(event)


@pytest.fixture(scope="session")
def fake_publisher():
    return FakePublisher()


@pytest.fixture(scope="session")
def in_memory_db():
    engine = create_engine("sqlite:///:memory:")
    metadata.create_all(engine)
    return engine


@pytest.fixture(scope="session")
def session(in_memory_db):
    start_mappers()
    yield sessionmaker(bind=in_memory_db)()
    clear_mappers()


@pytest.fixture(scope="session")
def uow():
    return unit_of_work.SqlAlchemyUnitOfWork("sqlite:///:memory:")


@pytest.fixture(scope="session")
def bus(fake_publisher, uow):
    return bootstrap.bootstrap(start_orm=True,
                               uow=uow,
                               publish_service=fake_publisher,
                               config={
                                   "user": "test_user",
                                   "workspace": "default"
                               })



# TODO: Move out of service layer
class TerkaException(Exception):
    ...


class TerkaCommandException(TerkaException):
    ...


class TaskAddedToCompletedEntity(TerkaException):
    ...


class TaskAddedToEntity(TerkaException):
    ...


class EntityNotFound(TerkaException):
    ...


class TerkaSprintEndDateInThePast(TerkaException):
    ...


class TerkaSprintCompleted(TerkaException):
    ...


class TerkaSprintActive(TerkaException):
    ...


class TerkaInitError(TerkaException):
    ...




from terka import utils

def test_create_task_dict():
    kwargs = ['-n', 'task_name', '-p', 'project_name', '-a', 'am']
    expected = {'n': 'task_name', 'p': 'project_name', 'a': 'am'}
    task_dict = utils.create_task_dict(kwargs)
    assert task_dict == expected



from __future__ import annotations
from typing import Dict, List, Optional
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
from sqlalchemy import text

from terka.domain import entities


def n_days_ago(n: int) -> str:
    return (datetime.today() - timedelta(n)).strftime("%Y-%m-%d")


def projects(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.project.Project)]


def projects_id_to_name_mapping(uow) -> dict[int, str]:
    with uow:
        return {t.id: t.name for t in uow.tasks.list(entities.project.Project)}


def project(uow, project_id: int):
    with uow:
        if not (project := uow.tasks.get_by_id(entities.project.Project,
                                               project_id)):
            return {}
        tasks = [task.to_dict() for task in project.tasks]
        result = project.to_dict()
        result["tasks"] = tasks
        return result


def tasks(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.task.Task)]


def task(uow, task_id: int):
    with uow:
        if not (task := uow.tasks.get_by_id(entities.task.Task, task_id)):
            return {}
        result = task.to_dict()
        if commentaries := task.commentaries:
            commentaries = [comment.to_dict() for comment in commentaries]
        result["commentaries"] = commentaries
        return result


def task_commentaries(uow, task_id: int) -> dict:
    with uow:
        if not (task := uow.tasks.get_by_id(entities.task.Task, task_id)):
            return {}
        if commentaries := task.commentaries:
            return [comment.to_dict() for comment in commentaries]
        return {}


def workspaces(uow) -> list[dict]:
    with uow:
        return [
            t.to_dict() for t in uow.tasks.list(entities.workspace.Workspace)
        ]


def workspaces_id_to_name_mapping(uow) -> dict[int, str]:
    with uow:
        return {
            t.id: t.name
            for t in uow.tasks.list(entities.workspace.Workspace)
        }


def workspace(uow, workspace_id: int):
    with uow:
        if not (workspace := uow.tasks.get_by_id(entities.workspace.Workspace,
                                                 workspace_id)):
            return {}
        if projects := workspace.projects:
            projects = [project.to_dict() for project in workspace.projects]
            result = workspace.to_dict()
        result["projects"] = projects
        return result


def workspace_projects(uow, workspace_id: int):
    with uow:
        if not (workspace := uow.tasks.get_by_id(entities.workspace.Workspace,
                                                 workspace_id)):
            return {}
        if projects := workspace.projects:
            return [project.to_dict() for project in workspace.projects]
        return {}


def epics(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.epic.Epic)]


def epic(uow, epic_id: int):
    with uow:
        if not (epic := uow.tasks.get_by_id(entities.epic.Epic, epic_id)):
            return {}
        if tasks := epic.tasks:
            tasks = [task.tasks.to_dict() for task in epic.tasks]
            result = epic.to_dict()
        result["tasks"] = tasks
        return result


def epic_tasks(uow, epic_id: int):
    with uow:
        if not (epic := uow.tasks.get_by_id(entities.epic.Epic, epic_id)):
            return {}
        if tasks := epic.tasks:
            return [task.tasks.to_dict() for task in epic.tasks]
        return {}


def stories(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.story.Story)]


def story(uow, story_id: int):
    with uow:
        if not (story := uow.tasks.get_by_id(entities.story.Story, story_id)):
            return {}
        if tasks := story.tasks:
            tasks = [task.tasks.to_dict() for task in story.tasks]
            result = story.to_dict()
        result["tasks"] = tasks
        return result


def story_tasks(uow, story_id: int):
    with uow:
        if not (story := uow.tasks.get_by_id(entities.story.Story, story_id)):
            return {}
        if tasks := story.tasks:
            return [task.tasks.to_dict() for task in story.tasks]
        return {}


def sprints(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.sprint.Sprint)]


def sprint(uow, sprint_id: int):
    with uow:
        if not (sprint := uow.tasks.get_by_id(entities.sprint.Sprint,
                                              sprint_id)):
            return {}
        if tasks := sprint.tasks:
            tasks = [task.tasks.to_dict() for task in sprint.tasks]
            result = sprint.to_dict()
        result["tasks"] = tasks
        return result


def sprint_tasks(uow, sprint_id: int):
    with uow:
        if not (sprint := uow.tasks.get_by_id(entities.sprint.Sprint,
                                              sprint_id)):
            return {}
        if tasks := sprint.tasks:
            return [task.tasks.to_dict() for task in sprint.tasks]
        return {}


def users(uow):
    with uow:
        return [t.to_dict() for t in uow.tasks.list(entities.user.User)]


def users_id_to_name_mapping(uow) -> dict:
    with uow:
        return {t.id: t.name for t in uow.tasks.list(entities.user.User)}


def user(uow, user_id: int):
    with uow:
        if not (user := uow.tasks.get_by_id(entities.user.User, user_id)):
            return {}
        return user.to_dict()


def time_spent(session,
               tasks: list[entities.task.Task] | None,
               start_date: str = n_days_ago(7),
               end_date: str = n_days_ago(-1),
               excluded_tasks_only: bool = False) -> List[Dict[str, str]]:
    tasks = ",".join([str(task.id) for task in tasks])
    query = f"""
    SELECT
        STRFTIME("%Y-%m-%d", creation_date) AS date,
        SUM(time_spent_minutes) AS time_spent_hours
    FROM time_tracker_entries
    WHERE
        creation_date >= "{start_date}"
        AND creation_date <= "{end_date}"
    """
    if excluded_tasks_only:
        query += f" AND task NOT IN ({tasks})"
    else:
        query += f" AND task IN ({tasks})"
    query += "GROUP BY 1"
    results = session.execute(text(query))
    return [dict(r) for r in results]


def sprint_task_ids(session,
                    sprint_id: Optional[int] = None) -> List[Dict[int, int]]:
    results = session.execute(
        """
    SELECT
        task, sprint
    FROM sprint_tasks
    WHERE
        sprint = :sprint_id
    """, dict(sprint_id=sprint_id))
    return [dict(r) for r in results]


def external_connectors_asana_projects(session) -> List[Dict[int, str]]:
    results = session.execute("""
    SELECT
        id, asana_project_id, sync_date
    FROM 'external_connectors.asana.projects'
    """)
    return [dict(r) for r in results]


def external_connectors_asana_tasks(
        session,
        project_id: str) -> Dict[int, Dict[str, str | datetime | None]]:
    results = session.execute(
        """
    SELECT
        id, asana_task_id, sync_date
    FROM 'external_connectors.asana.tasks'
    WHERE project = :project_id
    """, dict(project_id=project_id))
    return {
        r.id: {
            "asana_task_id":
            r.asana_task_id,
            "sync_date":
            datetime.strptime(r.sync_date, "%Y-%m-%d %H:%M:%S.%f")
            if r.sync_date else None
        }
        for r in results
    }


def status_changes(session, sprint_id: int):
    query = f"""
        SELECT DISTINCT T.id AS task,
        T.status AS current_status,
        STRFTIME("%Y-%m-%d", E.date) AS date,
        FIRST_VALUE(E.old_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS first_status_for_date,
        LAST_VALUE(E.new_value) OVER (PARTITION BY T.id, STRFTIME("%Y-%m-%d", E.date) ORDER by E.date RANGE BETWEEN UNBOUNDED PRECEDING AND
                UNBOUNDED FOLLOWING) AS last_status_for_date
        FROM tasks AS T
        LEFT JOIN (SELECT * FROM task_events WHERE type = "STATUS") AS E
            ON T.id = E.task
        WHERE T.id IN (SELECT task FROM sprint_tasks WHERE sprint = {sprint_id})

        """
    return pd.read_sql_query(query, session.get_bind()).replace({np.nan: None})



import inspect

from terka.adapters import orm, publisher
from terka.service_layer import handlers, messagebus, unit_of_work


def bootstrap(
    uow: unit_of_work.AbstractUnitOfWork,
    start_orm: bool = True,
    publish_service: publisher.BasePublisher = publisher.LogPublisher(),
    config: dict | None = None
) -> messagebus.MessageBus:

    if start_orm:
        orm.start_mappers(engine=uow.engine)

    return messagebus.MessageBus(uow=uow,
                                 publisher=publish_service,
                                 event_handlers=handlers.EVENT_HANDLERS,
                                 command_handlers=handlers.COMMAND_HANDLERS,
                                 config=config)



from datetime import datetime
from textual import on
from textual.app import ComposeResult
from textual.containers import Container, Grid
from textual.reactive import reactive
from textual.screen import ModalScreen
from textual.validation import Number
from textual.widgets import Button, Input, Label, Select, Static

from terka.domain import commands


class Text(Static):
    text = reactive("", layout=True)

    def render(self) -> str:
        return f"{self.__class__.__name__}: {self.text}"


class Value(Static):
    value = reactive("", layout=True)

    def render(self) -> str:
        return f"{self.__class__.__name__}: {self.value}"


class Values(Static):
    values = reactive([], layout=True)

    def render(self) -> str:
        return f"{self.__class__.__name__}: \n{[v for v in self.values]}"


class Title(Text):
    ...


class Description(Text):
    ...


class Status(Value):
    ...


class Priority(Value):
    ...


class Project(Value):
    ...


class Commentaries(Values):
    ...


class Sidebar(Container):

    def compose(self) -> ComposeResult:
        yield Container(Title(classes="header"), Description(), Status(),
                        Priority(), Project(), Commentaries())


class EpicSidebar(Container):

    def compose(self) -> ComposeResult:
        yield Container(Title(classes="header"), Description())


class StorySidebar(Container):

    def compose(self) -> ComposeResult:
        yield Container(Title(classes="header"), Description())


class TerkaModalScreen(ModalScreen[str]):
    BINDINGS = [("escape", "quit", "Quit")]

    @on(Input.Submitted)
    def submit_input(self, event: Input.Submitted) -> None:
        self.app.pop_screen()

    def action_quit(self):
        self.app.pop_screen()


class NewTask(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"New task", id="question"),
            Input(placeholder="Name", id="name", classes="text-input"),
            Input(placeholder="Description",
                  id="description",
                  classes="description-input"),
            Input(placeholder="Project", id="project",
                  classes="project-input"),
            Input(placeholder="Assignee",
                  id="assignee",
                  classes="assignee-input"),
            Select(((line, line) for line in [
                "BACKLOG", "TODO", "IN_PROGRESS", "REVIEW", "DONE", "DELETED"
            ]),
                   prompt="status",
                   value="BACKLOG",
                   id="status"),
            Select(
                ((line, line) for line in ["LOW", "NORMAL", "HIGH", "URGENT"]),
                prompt="priority",
                value="NORMAL",
                id="priority"),
            Input(placeholder="Add due date", id="due_date"),
            Input(placeholder="sprint", id="sprint", validators=[Number()]),
            Input(placeholder="epic", id="epic", validators=[Number()]),
            Input(placeholder="story", id="story", validators=[Number()]),
            Input(placeholder="tags", id="tags"),
            Input(placeholder="collaborators", id="collaborators"),
            Button("Save", id="yes", variant="success"),
            Button("No", id="no"),
            id="new-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            name = self.query_one("#name", Input)
            description = self.query_one("#description", Input)
            project = self.query_one("#project", Input)
            status = self.query_one("#status", Select)
            priority = self.query_one("#priority", Select)
            due_date = self.query_one("#due_date", Input)
            if due_date := due_date.value:
                due_date = datetime.strptime(due_date, "%Y-%m-%d")
            else:
                due_date = None
            sprint = self.query_one("#sprint", Input)
            epic = self.query_one("#epic", Input)
            story = self.query_one("#story", Input)
            tags = self.query_one("#tags", Input)
            collaborators = self.query_one("#collaborators", Input)
            returned_commands = []
            create_command = commands.CreateTask(name=name.value,
                                                 description=description.value,
                                                 project=project.value,
                                                 due_date=due_date,
                                                 status=status.value,
                                                 priority=priority.value)
            returned_commands.append(create_command)
            if sprint.value or story.value or epic.value:
                add_command = commands.AddTask(id=None,
                                               sprint=sprint.value,
                                               epic=epic.value,
                                               story=story.value)
                returned_commands.append(add_command)
            if tags.value:
                tag_command = commands.TagTask(id=None, tag=tags.value)
                returned_commands.append(tag_command)
            if collaborators.value:
                collaborator_command = commands.CollaborateTask(
                    id=None, collaborator=collaborators.value)
                returned_commands.append(collaborator_command)
            self.dismiss(returned_commands)
        else:
            self.app.pop_screen()


class NewEpic(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"New epic", id="question"),
            Input(placeholder="Name", id="name", classes="text-input"),
            Input(placeholder="Description",
                  id="description",
                  classes="description-input"),
            Button("Save", id="yes", variant="success"),
            Button("No", id="no"),
            id="new-composite-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            name = self.query_one("#name", Input)
            description = self.query_one("#description", Input)
            self.dismiss(
                commands.CreateEpic(name=name.value,
                                    description=description.value))
        else:
            self.app.pop_screen()


class NewStory(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"New story", id="question"),
            Input(placeholder="Name", id="name", classes="text-input"),
            Input(placeholder="Description",
                  id="description",
                  classes="description-input"),
            Button("Save", id="yes", variant="success"),
            Button("No", id="no"),
            id="new-composite-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            name = self.query_one("#name", Input)
            description = self.query_one("#description", Input)
            self.dismiss(
                commands.CreateStory(name=name.value,
                                     description=description.value))
        else:
            self.app.pop_screen()


class TaskAdd(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Add task", id="question"),
            Input(placeholder="To sprint", id="sprint", validators=[Number()]),
            Input(placeholder="To epic", id="epic", validators=[Number()]),
            Input(placeholder="To story", id="story", validators=[Number()]),
            Input(placeholder="Story points",
                  id="story-points",
                  validators=[Number()]),
            Button("Confirm", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            sprint = self.query_one("#sprint", Input)
            epic = self.query_one("#epic", Input)
            story = self.query_one("#story", Input)
            story_points = self.query_one("#story-points", Input)
            self.dismiss(
                commands.AddTask(id=None,
                                 sprint=sprint.value,
                                 epic=epic.value,
                                 story=story.value,
                                 story_points=story_points.value))
        else:
            self.app.pop_screen()


class ShowTask(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label("Task info", id="question"),
            Static(f"[b]Name:[/b] ", id="name"),
            Static("Description", id="description"),
            Static("status", id="status"),
            Static("priority", id="priority"),
            Static("due date", id="due_date"),
            id="edit-dialog",
        )


class TaskEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Edit task", id="question"),
            Input(placeholder="Name", id="name"),
            Input(placeholder="Description",
                  id="description",
                  classes="description-input"),
            Select(((line, line) for line in [
                "BACKLOG", "TODO", "IN_PROGRESS", "REVIEW", "DONE", "DELETED"
            ]),
                   prompt="status",
                   id="status"),
            Select(
                ((line, line) for line in ["LOW", "NORMAL", "HIGH", "URGENT"]),
                prompt="priority",
                id="priority"),
            Input(placeholder="Add due date", id="due_date"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Input(placeholder="Comment", id="comment"),
            Button("Save", id="yes", variant="success"),
            Button("No", id="no"),
            id="edit-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            name = self.query_one("#name", Input)
            description = self.query_one("#description", Input)
            status = self.query_one("#status", Select)
            priority = self.query_one("#priority", Select)
            due_date = self.query_one("#due_date", Input)
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            if due_date := due_date.value:
                due_date = datetime.strptime(due_date, "%Y-%m-%d")
            commands_chain = []
            update_cmd = commands.UpdateTask(
                id=None,
                name=name.value,
                description=description.value,
                # FIXME: Bring back
                # due_date=due_date,
                status=status.value,
                priority=priority.value)
            if update_cmd:
                commands_chain.append(update_cmd)
            commands_chain.append(
                commands.CommentTask(id=None, text=comment.value))
            commands_chain.append(
                commands.TrackTask(id=None, hours=hours.value))

            self.dismiss(tuple(commands_chain))
        else:
            self.app.pop_screen()


class TaskComplete(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Complete task", id="question"),
            Input(placeholder="Add comment", id="comment"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Button("Yes", id="yes", variant="success"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            self.dismiss((commands.CompleteTask(id=None),
                          commands.CommentTask(id=None, text=comment.value),
                          commands.TrackTask(id=None, hours=hours.value)))
        else:
            self.app.pop_screen()


class TaskDelete(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Delete task", id="question"),
            Input(placeholder="Add comment", id="comment"),
            Input(placeholder="Add time spent",
                  validators=[Number()],
                  id="hours"),
            Button("Yes", id="yes", variant="success"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            comment = self.query_one("#comment", Input)
            hours = self.query_one("#hours", Input)
            self.dismiss((commands.DeleteTask(id=None),
                          commands.CommentTask(id=None, text=comment.value),
                          commands.TrackTask(id=None, hours=hours.value)))
        else:
            self.app.pop_screen()


class TaskStatusEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Change status", id="question"),
            Select(((line, line) for line in [
                "BACKLOG", "TODO", "IN_PROGRESS", "REVIEW", "DONE", "DELETED"
            ]),
                   classes="small-select",
                   prompt="status",
                   id="status"),
            Button("Confim", id="yes", variant="success"),
            Button("Cancel", id="no"),
            id="small-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            status = self.query_one("#status", Select)
            self.dismiss(status.value)
        else:
            self.app.pop_screen()


class TaskPriorityEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Change priority", id="question"),
            Select(
                ((line, line) for line in ["LOW", "NORMAL", "HIGH", "URGENT"]),
                prompt="priority",
                classes="small-select",
                id="priority"),
            Button("Confim", id="yes", variant="success"),
            Button("Cancel", id="no"),
            id="small-dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            priority = self.query_one("#priority", Select)
            self.dismiss(priority.value)
        else:
            self.app.pop_screen()


class TaskStoryPointsEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Add story points", id="question"),
            Input(placeholder="Story points",
                  id="story-points",
                  validators=[Number()]),
            Button("Confim", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            story_points = self.query_one("#story-points", Input)
            self.dismiss(story_points.value)
        else:
            self.app.pop_screen()


class TaskHoursSubmitted(ModalScreen[str]):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Add time spent", id="question"),
            Input(placeholder="Time spent in minutes",
                  id="hours",
                  validators=[Number()]),
            Button("Confim", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            story_points = self.query_one("#hours", Input)
            self.dismiss(story_points.value)
        else:
            self.app.pop_screen()


class TaskTagEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Add tag", id="question"),
            Input(placeholder="Tags", id="tags"),
            Button("Confim", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            tags = self.query_one("#tags", Input)
            self.dismiss(tags.value)
        else:
            self.app.pop_screen()


class TaskCollaboratorEdit(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Add collaborators", id="question"),
            Input(placeholder="Collaborators", id="collaborators"),
            Button("Confim", id="yes"),
            Button("Cancel", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            collaborators = self.query_one("#collaborators", Input)
            self.dismiss(collaborators.value)
        else:
            self.app.pop_screen()


class TaskDeleteDueDate(TerkaModalScreen):

    def compose(self) -> ComposeResult:
        yield Grid(
            Label(f"Remove due date", id="question"),
            Button("Yes", id="yes"),
            Button("No", id="no"),
            id="dialog",
        )

    def on_button_pressed(self, event: Button.Pressed) -> None:
        if event.button.id == "yes":
            self.dismiss(True)
        else:
            self.app.pop_screen()



from typing import Set
from collections import defaultdict
from enum import Enum
import logging
from datetime import datetime, date
from dataclasses import dataclass

from .entity import Entity
from .task import Task

logger = logging.getLogger(__name__)


class SprintStatus(Enum):
    PLANNED = 1
    ACTIVE = 2
    COMPLETED = 3
    DELETED = 4


class Sprint(Entity):

    def __init__(self,
                 start_date: datetime = None,
                 end_date: datetime = None,
                 status: str = "PLANNED",
                 goal: str = None,
                 **kwargs) -> None:
        if not start_date and not end_date:
            raise ValueError("Please add start and end date of the sprint")
        if start_date.date() < datetime.today().date():
            raise ValueError(f"start date cannot be less than today")
        if not start_date:
            raise ValueError(
                "Please provide start date for the sprint in YYYY-MM-DD format"
            )
        self.start_date = start_date
        if not end_date:
            raise ValueError(
                "Please provide end date for the sprint in YYYY-MM-DD format")
        self.end_date = end_date
        if end_date.date() < start_date.date():
            raise ValueError(f"Sprint end date cannot be less than start date")
        if end_date.date() < datetime.today().date():
            raise ValueError(f"Sprint end date cannot be less than today")
        self.status = self._validate_status(status)
        self.goal = goal
        self.is_completed = False

    def _validate_status(self, status):
        if status and status not in [s.name for s in SprintStatus]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    def complete(self, tasks) -> None:
        incompleted_tasks = list()
        for task in tasks:
            if task.tasks.status.name != "DONE":
                incompleted_tasks.append(task)
        if incompleted_tasks:
            logging.warning("[Sprint %d]: %d tasks haven't been completed",
                            self.id, len(incompleted_tasks))
        self.is_completed = True

    @property
    def velocity(self) -> float:
        return round(sum([t.story_points for t in self.tasks]), 1)

    @property
    def utilization(self) -> float:
        if not (velocity := self.velocity) or not (total_time_spent :=
                                                   self.total_time_spent):
            return 0
        return total_time_spent / (velocity * 60)

    @property
    def time_spent_today(self):
        time_spent_today = 0
        for sprint_task in self.tasks:
            time_spent_today += sprint_task.tasks.time_spent_today
        return time_spent_today

    @property
    def total_time_spent(self):
        total_time_spent_sprint = 0
        for sprint_task in self.tasks:
            total_time_spent_sprint += sprint_task.tasks.total_time_spent
        return total_time_spent_sprint

    @property
    def open_tasks(self):
        return [
            task for task in self.tasks
            if task.tasks.status.name not in ("DONE", "DELETED")
        ]

    @property
    def pct_completed(self) -> float:
        if (total_tasks := len(self.tasks)) > 0:
            return (total_tasks - len(self.open_tasks)) / total_tasks
        return 0

    @property
    def collaborators(self):
        collaborators = defaultdict(int)
        for sprint_task in self.tasks:
            if task_collaborators := sprint_task.tasks.collaborators:
                for collaborator in task_collaborators:
                    name = collaborator.users.name or "me"
                    collaborators[name] += sprint_task.tasks.total_time_spent
            else:
                collaborators["me"] += sprint_task.tasks.total_time_spent
        return collaborators

    @property
    def collaborators_as_string(self):
        collaborators = []
        for user, story_point in sorted(self.collaborators.items(),
                                        key=lambda x: x[1],
                                        reverse=True):
            collaborators.append(f"{user} ({round(story_point, 2)})")
        return ",".join(collaborators)

    def daily_time_entries_hours(self) -> dict[str, float]:
        entries: dict[str, float] = defaultdict(float)
        for task in self.tasks:
            task_entries = task.tasks.daily_time_entries_hours(
                self.start_date, self.end_date)
            for day, hours in task_entries.items():
                entries[day] += hours
        return entries


@dataclass
class SprintTask:
    sprint: int
    task: int
    story_points: int = 0
    #TODO: add actual_time_spent: int = 0
    is_active_link: bool = True



import abc
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from terka.adapters import repository
from terka.domain import events


class AbstractUnitOfWork(abc.ABC):
    tasks: repository.AbsRepository
    published_events: list[events.Event] = []

    def __enter__(self):
        return self

    def __exit__(self, *args):
        self.rollback()

    def commit(self):
        self._commit()

    def flush(self):
        self._flush()

    def collect_new_events(self):
        while self.published_events:
            yield self.published_events.pop(0)

    @abc.abstractmethod
    def _commit(self):
        ...

    @abc.abstractmethod
    def _flush(self):
        ...

    @abc.abstractmethod
    def rollback(self):
        ...


class SqlAlchemyUnitOfWork(AbstractUnitOfWork):

    def __init__(self, session_factory) -> None:
        self.engine = create_engine(session_factory)
        self.session_factory = sessionmaker(self.engine)
        self.published_events: list[events.Event] = []

    @property
    def repo(self):
        return repository.SqlAlchemyRepository(self.session_factory())

    def __enter__(self) -> None:
        self.session = self.session_factory()  # type: Session
        self.tasks = repository.SqlAlchemyRepository(self.session)
        return super().__enter__()

    def __exit__(self, *args):
        super().__exit__(*args)
        self.session.close()

    def _commit(self):
        self.session.commit()

    def _flush(self):
        self.session.flush()

    def rollback(self):
        self.session.rollback()



from datetime import datetime

from .entity import Entity


class BaseCommentary(Entity):

    def __init__(self,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.text = text
        self.date = date

    def __repr__(self):
        return f"<Commentary> {self.text}"

    def __bool__(self) -> bool:
        return bool(self.text)


class TaskCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.task = id
        super().__init__(text=text, date=date)


class ProjectCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.project = id
        super().__init__(text=text, date=date)


class EpicCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.epic = id
        super().__init__(text=text, date=date)


class StoryCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.story = id
        super().__init__(text=text, date=date)


class SprintCommentary(BaseCommentary):

    def __init__(self,
                 id: int,
                 text: str,
                 date: datetime = datetime.now(),
                 **kwargs: str) -> None:
        self.sprint = id
        super().__init__(text=text, date=date)



from collections import defaultdict
from datetime import date, datetime
from enum import Enum
from statistics import median

from .entity import Entity

class ProjectStatus(Enum):
    DELETED = 0
    ACTIVE = 1
    ON_HOLD = 2
    COMPLETED = 3


class Project(Entity):

    _next_project_id = 1

    def __init__(self,
                 name: str,
                 description: str | None = None,
                 created_by: str | None = None,
                 status: str = "ACTIVE",
                 workspace: int | None = None) -> None:
        self._project_id = Project._next_project_id
        Project._next_project_id += 1
        self.name = name
        self.description = description
        self.created_by = created_by
        self.status = status
        self.workspace = workspace

    def _validate_status(self, status):
        if status not in [
                s.name for s in ProjectStatus if s.name != "DELETED"
        ]:
            raise ValueError(f"{status} is invalid status")
        else:
            return status

    @property
    def total_time_spent(self) -> int:
        total_time_spent_project = 0
        for task in self.tasks:
            total_time_spent_project += task.total_time_spent
        return total_time_spent_project

    @property
    def task_collaborators(self) -> dict[str, int]:
        collaborators = defaultdict(int)
        for task in self.tasks:
            if task_collaborators := task.collaborators:
                for collaborator in task.collaborators:
                    name = collaborator.users.name or "me"
                    collaborators[name] += task.total_time_spent
            else:
                collaborators["me"] += task.total_time_spent
        return collaborators

    @property
    def open_tasks(self):
        return [
            task for task in self.tasks
            if task.status.name not in ("DONE", "DELETED")
        ]

    @property
    def closed_tasks(self):
        return [
            task for task in self.tasks
            if task.status.name in ("DONE", "DELETED")
        ]

    @property
    def overdue_tasks(self):
        return [
            task for task in self.open_tasks
            if task.due_date and task.due_date <= datetime.now().date()
        ]

    @property
    def stale_tasks(self):
        return [task for task in self.open_tasks if task.is_stale]

    @property
    def median_task_age(self):
        if self.open_tasks:
            return round(
                median([(datetime.now() - task.creation_date).days
                        for task in self.open_tasks]))
        return 0

    @property
    def backlog(self):
        return self._count_task_status("BACKLOG")

    @property
    def todo(self):
        return self._count_task_status("TODO")

    @property
    def in_progress(self):
        return self._count_task_status("IN_PROGRESS")

    @property
    def review(self):
        return self._count_task_status("REVIEW")

    @property
    def done(self):
        return self._count_task_status("DONE")

    @property
    def deleted(self):
        return self._count_task_status("DELETED")

    def daily_time_entries_hours(
            self,
            start_date: str | date | None = None,
            end_date: str | date | None = None,
            last_n_days: int | None = None) -> dict[str, float]:
        entries: dict[str, float] = defaultdict(float)
        for task in self.tasks:
            task_entries = task.daily_time_entries_hours(
                start_date, end_date, last_n_days)
            for day, hours in task_entries.items():
                entries[day] += hours
        return entries

    def __str__(self):
        return f"<Project {self.id}>: {self.name} {self.tasks}"

    def _count_task_status(self, status: str) -> int:
        return [task for task in self.tasks if task.status.name == status]



from __future__ import annotations
import logging
from pathlib import Path
import rich
from rich.console import Console
from rich.table import Table
import os
import re
import yaml

from terka.adapters.repository import AbsRepository
from terka.domain import entities


def update_config(update_pair: dict):
    config = get_config()
    config.update(update_pair)
    with open(f"{home_dir}/.terka/config.yaml", "w") as f:
        yaml.dump(config, f)


def get_config():
    home_dir = os.path.expanduser('~')
    with open(f"{home_dir}/.terka/config.yaml", "r", encoding="utf-8") as f:
        config = yaml.safe_load(f)
    return config


def get_workplace_by_name(
        workspace_name: str,
        repo: AbsRepository) -> entities.workspace.Workspace | None:
    return repo.get(entities.workspace.Workspace, workspace_name)


class ServiceCommandHander:

    def __init__(self, home_dir, config, console=Console()):
        self.home_dir = home_dir
        self.config = config
        self.console = console

    def execute(self, command, entity, task_dict):
        is_service_command = True
        if command == "init":
            self.init_terka()
        elif command == "focus":
            self.focus(entity, task_dict)
        elif command == "unfocus":
            self.unfocus()
        elif command == "switch" and entity == "workspace":
            self.switch(task_dict.get("id"))
        elif command == "set" and entity == "workspace":
            self.set(task_dict.get("id"))
        elif command == "log":
            self.log(task_dict)
        else:
            is_service_command = False
        if is_service_command:
            exit()

    def init_terka(self,
                   terka_folder: str = ".terka",
                   default_user: str = "admin",
                   default_config: str = "config.yaml") -> None:
        path = os.path.join(self.home_dir, terka_folder)
        if not os.path.exists(path):
            answer = input(
                f"Do you want to init terka in this directory {path}? [Y/n]")
            if "y" in answer.lower():
                Path(path).mkdir(parents=True, exist_ok=True)
                with open(os.path.join(path, "config.yaml"), "w") as f:
                    yaml.dump({"user": default_user}, f)
            elif "n" in answer.lower():
                path = input("Specify full path to the terka directory: ")
                os.mkdirs(path)
            else:
                exit()
        elif not os.path.exists(os.path.join(path, default_config)):
            answer = input(
                f"Config.yaml not found in {path}, Create it? [Y/n]")
            if "y" in answer.lower():
                with open(os.path.join(path, default_config), "w") as f:
                    yaml.dump({"user": default_user}, f)
            else:
                exit()
        else:
            print("Terka directory already exist.")

    def focus(self, entity_type, kwargs):
        if entity_type == "tasks":
            self.config["task_id"] = kwargs["id"]
            if "project_name" in self.config.keys():
                del self.config["project_name"]
        if entity_type == "projects":
            self.config["project_name"] = kwargs["id"]
            if "task_id" in self.config.keys():
                del self.config["task_id"]
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("<focus> %s: %s", entity_type, "")

    def unfocus(self):
        if "task_id" in self.config.keys():
            del self.config["task_id"]
        if "project_name" in self.config.keys():
            del self.config["project_name"]
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        loging.info("<unfocus> %s: %s", "", "")

    def log(self, kwargs):
        table = Table(box=rich.box.SIMPLE)
        with open(f"{self.home_dir}/.terka/terka.log", "r") as f:
            head = f.readlines()
        for column in ("date", "source", "level", "message"):
            table.add_column(column)
        num_log_entries = int(kwargs.get("num_log_entries", 10))
        tail = head[-num_log_entries:]
        for row in tail[::-1]:
            info, message = row.split("] ")
            date, source, level = info.split("][")
            table.add_row(re.sub("\[", "", date), source, level, message)
        self.console.print(table)

    def switch(self, workspace: str):
        if not self.config.get("workspace"):
            self.config["workspace"] = "Default"  # Default workspace
        else:
            self.config["workspace"] = workspace
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("switched to workspace: %d", workspace)

    def set_workspace(self, workspace: str):
        self.config["workspace"] = workspace
        with open(f"{self.home_dir}/.terka/config.yaml", "w",
                  encoding="utf-8") as f:
            yaml.dump(self.config, f)
        logging.info("set workspace: %d", workspace)
